from pydantic import BaseModel
from bson import ObjectId
from enum import Enum
from datetime import datetime


class DataTypes(Enum):
    STRING = str
    INT = int
    FLOAT = float
    BOOLEAN = bool
    DATE = datetime.now
    OBJECTID = ObjectId
    ARRAY = list
    OBJECT = dict
    NULL = None


class Schema(BaseModel):
    name: str
    fields: set[str] = set()


class Query(BaseModel):
    query: list = []
    fields: dict = {}
    empty: bool = True



import sys, asyncio
from dotenv import load_dotenv
from pathlib import Path
from .core import create_file_ini, get_config
from .repository.collections.models import Query
from .services.migration import (
    FileMigration, get_collections, connect_to_mongo, get_database,
    add_fields, delete_fields
)

load_dotenv()
PATH = Path(__file__).parent.resolve()


async def update_migration(client):
    collections_migration = FileMigration.get_migration()
    db = get_database(client)
    for collection in get_collections(client):
        if collection.name in collections_migration:
            fields = set(collections_migration[collection.name].keys())
            #Симметрическая разность множеств
            disjunctive = fields ^ collection.fields
            if "_id" in disjunctive:
                disjunctive.remove("_id")
            delete = Query()
            add = Query()
            for field in disjunctive:
                if field in fields:
                    add.query.append({field: {"$exists": False}})
                    add.fields[field] = collections_migration[collection.name][field]
                    add.empty = False
                if field not in fields and collection.fields:
                    delete.query.append({field: {"$exists": True}})
                    delete.fields[field] = ""
                    delete.empty = False
            add_fields(db, collection.name, add)
            delete_fields(db, collection.name, delete)


async def create_migration():
    config = get_config()
    module_path = config.get("MONGO", "module_path")
    migration_file = await FileMigration.make_migration(module_path)
    FileMigration.update_migration_file(migration_file)


async def main():
    FileMigration.create_migration_catalog()
    params = tuple(sys.argv[1:])
    # module_path = "app/src/app/schemes"
    # module_path = "utils/schemas/db_schemas_q_service"
    client = None
    try:
        if params[0] == "--create-migration":
            await create_migration()
        elif params[0] == "--update-migration":
            client = connect_to_mongo()
            await update_migration(client)
        elif params[0] == "--init":
            create_file_ini()
    finally:
        if client != None:
            client.close()
    # for collection in db_collections:
    #     if collection.name == "text_queue":
    #         for field in collection.fields:
    #             print(collection.name, field.name, field.type)
    # find_collection = list(filter(lambda x: x.name in {"customer_request"}, db_collections))
    # query = {
    #     "$or": [{field.name: {"$exists": False}} for field in find_collection[0].fields]
    # }
    # set_field = {field.name for field in find_collection[0].fields}
    # get_collection = db.get_collection("customer_request").find(query)
    # for item in get_collection:
    #     fields = set_field ^ set(item.keys())
    #     update_fields = {"$set": {field: None for field in fields}}
    #     db.get_collection("customer_request").update_many(
    #         query, update_fields
    #     )

if __name__ == "__main__":
    asyncio.run(main())


import pymongo, logging
from pymongo.database import Database
from mongo_m.core import MongoDB, get_config
from mongo_m.repository import collections

__all__ = ['connect_to_mongo', 'get_collections', 'get_database', 'delete_fields', 'add_fields']


def connect_to_mongo() -> pymongo.MongoClient:
    """
    Connects to a MongoDB database using environment variables for configuration.

    Returns:
    - pymongo.database.Database: The connected MongoDB database.
    """
    config = get_config()
    return MongoDB(config.get("MONGO", "host"),
                   int(config.get("MONGO", "port")),
                   config.get("MONGO", "user"),
                   config.get("MONGO", "password"))


def get_collections(client: pymongo.MongoClient):
    db = get_database(client)
    return collections.get_fields_collections(db)


def get_database(client: pymongo.MongoClient):
    config = get_config()
    db_name = config.get("MONGO", "database")
    db = collections.find_collections(client, db_name)
    return db


def delete_fields(db: Database, collection_name: str, params):
    if params.empty:
        return
    collections = db.get_collection(collection_name)
    try:
        collections.update_many({"$or": params.query}, {"$unset": params.fields}, upsert=True)
    except Exception as e:
        print(e)


def add_fields(db: Database, collection_name: str, params):
    if params.empty:
        return
    collections = db.get_collection(collection_name)
    try:
        collections.update_many({"$or": params.query}, {"$set": params.fields}, upsert=True)
    except Exception as e:
        print(e)


import configparser, os

__all__ = ['create_file_ini', 'get_config']

FILE = f"{os.getcwd()}/migration/config.ini"

def create_file_ini():
    config = configparser.ConfigParser()
    config['MONGO'] = {
        'host': 'localhost',
        'port': 27020,
        'user': 'devUser',
        'password': 'nbgfre736251',
        'database': 'queue_service',
        'module_path': 'migration'
    }
    with open(FILE, 'w') as f:
        config.write(f)

def get_config() -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    config.read(FILE)
    return config


"""
    Создание и удаление директорий программы и файлов
"""

import os

__all__ = ["check_catalog", "create_catalog"]


def check_catalog(catalog: str) -> bool:
    """
    Проверка существования директории
    """
    return os.path.exists(catalog)


def create_catalog(catalog: str) -> None:
    """
    Создание директории
    """
    if not check_catalog(catalog):
        os.mkdir(catalog)



import os, sys, importlib, inspect
from pathlib import Path
from .utils import get_default_value

__all__ = ["make_module", "inspect_module", "get_module"]

async def make_module(module_path: str):
    path = f"{os.getcwd()}/{module_path}".replace("\\", "/")
    print(path, os.path.isdir(path))
    if os.path.isdir(path):
        files = make_module_dir(path, module_path)
    else:
        path_module = module_path.replace("/", ".")
        files = [path_module]
    for name in files:
        try:
            module = get_module(name)
            yield module
        except ModuleNotFoundError as e:
            print(e)
            continue
    return

def make_module_dir(path: str, module_path: str):
    path_module = Path(path)
    module_path = module_path.replace("/", ".")
    result = []
    for file in filter(lambda x: x not in {"__init__.py", "__pycache__"}, os.listdir(path_module)):
        module = f"{module_path}.{file.replace('.py', '')}"
        result.append(module)
    return result

def get_module(module_path: str):
    if module_path in sys.modules:
        return sys.modules[module_path]
    return importlib.import_module(module_path)

def inspect_module(module) -> dict:
    """
    Состоявляем поля модели на основе класса
    """
    models = {}
    for name, obj in inspect.getmembers_static(module, inspect.isclass):
        if '__table_name__' in dir(obj):
            if obj.__table_name__ in models:
                continue
            params = {}
            for values in inspect.signature(obj).parameters.values():
                if values.name == 'id':
                    continue
                # Определение параметров по дефолту
                if not inspect.isclass(values.annotation):
                    """
                        Парсинг типов данных и установка значиний по умолчанию для создания схемы
                        пример 'bool | None' or 'int | str | None'
                        Берется первый элемент из массива
                    """
                    annotated_types = str(values.annotation).split("|")
                    annotated_types = annotated_types[0].strip()
                else:
                    annotated_types = values.annotation.__name__

                if values.default is inspect._empty:
                    default_value = get_default_value(annotated_types)
                else:
                    default_value = values.default

                params[values.name] = default_value
            models[obj.__table_name__] = params
    return models


def get_fields() -> list:
    """
    Получение полей с типами из коллекции
    """
    return [
        {
            "$project": {
                "fields": {
                    "$objectToArray": "$$ROOT"
                }
            }
        },
        {
            "$unwind": "$fields"
        },
        {
            "$group": {
                "_id": "$fields.k",
                "field": {"$first": "$fields.k"},
                "type": {
                    "$first":
                        {
                            "$cond": [
                                {"$eq": [{"$type": "$fields.v"}, "boolean"]},
                                {"$type": "$fields.v"},
                                "$fields.v"
                            ]
                        }
                }
            }
        },
        {"$match": {"type": {"$ne": None}}}
    ]



import sys
from setuptools import setup, find_packages

install_requires = [
    'annotated-types>=0.6.0',
    'configparser>=6.0.1',
    'dnspython>=2.6.1',
    'pydantic>=2.6.4',
    'pydantic-settings>=2.2.1',
    'pydantic_core>=2.16.3',
    'pymongo>=4.6.3',
    'python-dotenv>=1.0.1',
    'typing_extensions>=4.10.0'
]

DESCRIPTION = "A database migration tool for MongoDB"

def get_read_me():
    try:
        with open('README.md') as f:
            return f.read()
    except IOError:
        return DESCRIPTION

setup(
    name='mongo_m',
    version='1.0.4',
    description=DESCRIPTION,
    long_description=get_read_me(),
    long_description_content_type="text/markdown",
    author='Dankov Sergey',
    author_email='beyond31@mail.ru',
    license='GNU General Public License v3 (GPLv3)',
    url='https://github.com/BeyonD311/mongo-m',
    packages=find_packages(),
    include_package_data=True,
    platforms='any',
    zip_safe=False,
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)


from datetime import datetime

__all__ = ['get_default_value']

types = {
    'str': "",
    'int': 0,
    'list': [],
    'dict': {},
    'bool': False,
    'float': 0.0,
    'datetime': datetime.now(),
    'set': set(),
    'tuple': tuple()
}

def get_default_value(typing: str):
    typing = typing.lower()
    for type_key in types:
        if typing.find(type_key) != -1:
            return types[type_key]
    return None




from pymongo import MongoClient
from pymongo.database import Database, Collection
from .models import Schema
from .query import get_fields
from typing import Generator, Generic

__all__ = ["find_collections", "get_fields_collections"]


def find_collections(client: MongoClient, db_name: str) -> Database:
    db = client[db_name]
    return db


def get_fields_collections(db: Database) -> Generator[Schema, None, None]:
    query_field = get_fields()
    for collection_name in db.list_collection_names():
        collection = db.get_collection(collection_name)
        batches = list(collection.aggregate(query_field))
        collection_data = Schema(name=collection_name)
        [collection_data.fields.add(batch['field']) for batch in batches]
        yield collection_data




import os, json, hashlib
from pathlib import Path
from mongo_m.core.inspect_module import make_module, inspect_module
from mongo_m.core import check_catalog, create_catalog

__all__ = ['FileMigration']


class FileMigration:

    PATH_MIGRATION = Path(f"{os.getcwd()}/migration")
    PATH_JSON = Path(f"{os.getcwd()}/migration/update.json")

    @classmethod
    def create_migration_catalog(cls):
        if not check_catalog(cls.PATH_MIGRATION.name):
            create_catalog(cls.PATH_MIGRATION.name)

    @classmethod
    async def make_migration(cls, module_path: str):
        modules = []
        async for module in make_module(module_path):
            module = inspect_module(module)
            if module != {}:
                modules.append(module)
        modules = json.dumps(modules, indent=4)
        hash_name = hashlib.md5(modules.encode())
        file_name = f"{hash_name.hexdigest()}.json"
        with open(file=f"{cls.PATH_MIGRATION}/{file_name}", mode='w') as f:
            f.write(modules)
        return file_name

    @classmethod
    def update_migration_file(cls, file_migration: str):
        """
        Обновляет стек выполненых миграций
        """
        try:
            with open(file=cls.PATH_JSON, mode="r+") as f:
                file_data = f.read()
                items = json.loads(file_data)

                if len(items) == 0:
                    items.append(file_migration)
                elif file_migration != items[-1]:
                    items.append(file_migration)

                f.truncate(0)
                f.seek(0)
                f.write(json.dumps(items, indent=4))
                f.seekable()

        except FileNotFoundError:
            with open(file=cls.PATH_JSON, mode="w") as f:
                f.write(json.dumps([file_migration], indent=4))

    @classmethod
    def get_last_migration(cls):
        with open(file=cls.PATH_JSON, mode="r") as f:
            file_data = f.read()
            items = json.loads(file_data)
            if len(items) == 0:
                raise FileNotFoundError("File migration not found")
            return items[-1]

    @classmethod
    def get_migration(cls):
        last_migration = cls.get_last_migration()
        result = {}
        with open(file=f"{cls.PATH_MIGRATION}/{last_migration}", mode="r") as f:
            file_data = f.read()
            items = json.loads(file_data)
            for item in items:
                key = list(item.keys())[0]
                result[key] = item[key]
        return result


import pymongo


class MongoDB(pymongo.MongoClient):
    """
    Initializes a new MongoDB connection.

    Parameters:
        host (str): The host address of the MongoDB server.
        port (int): The port number of the MongoDB server.
        user (str, optional): The username for authentication. Defaults to None.
        password (str, optional): The password for authentication. Defaults to None.
    """

    # MongoDB connection
    def __init__(self, host, port, user: str = None, password: str = None):
        if user is not None and password is not None:
            uri = f"mongodb://{user}:{password}@{host}:{port}/"
        else:
            uri = f"mongodb://{host}:{port}/"
        super().__init__(uri, authSource="admin")
